'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

var _typeof2 = require('babel-runtime/helpers/typeof');

var _typeof3 = _interopRequireDefault(_typeof2);

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _getIterator2 = require('babel-runtime/core-js/get-iterator');

var _getIterator3 = _interopRequireDefault(_getIterator2);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _assign = require('babel-runtime/core-js/object/assign');

var _assign2 = _interopRequireDefault(_assign);

exports.default = subscribe;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _model = require('./model');

var _model2 = _interopRequireDefault(_model);

var _Doc = require('./types/Doc');

var _Doc2 = _interopRequireDefault(_Doc);

var _Query = require('./types/Query');

var _Query2 = _interopRequireDefault(_Query);

var _QueryExtra = require('./types/QueryExtra');

var _QueryExtra2 = _interopRequireDefault(_QueryExtra);

var _Local = require('./types/Local');

var _Local2 = _interopRequireDefault(_Local);

var _observerUtil = require('@nx-js/observer-util');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var DEFAULT_COLLECTION = '$components';
var SUBSCRIBE_COMPUTATION_NAME = '__subscribeComputation';
var HELPER_METHODS_TO_BIND = ['get', 'at', 'atMap', 'atForEach'];
var DUMMY_STATE = {};

function subscribe() {
  for (var _len = arguments.length, fns = Array(_len), _key = 0; _key < _len; _key++) {
    fns[_key] = arguments[_key];
  }

  return function derorateTarget(Component) {
    (0, _assign2.default)(Component.prototype, subscribeLocalMixin(fns, Component.prototype.componentWillMount, Component.prototype.componentWillUnmount, Component.prototype.componentWillReceiveProps));
    return Component;
  };
}

var subscribeLocalMixin = function subscribeLocalMixin(fns, oldComponentWillMount, oldComponentWillUnmount, oldComponentWillReceiveProps) {
  return {
    componentWillMount: function componentWillMount() {
      this.model = generateScopedModel();
      // also alias scoped model as $scope to be consistent
      // with the ${key} models for subscription items
      this.$scope = this.model;
      this.model.set('', (0, _observerUtil.observable)({})); // Initially set empty object for observable
      this.scope = this.model.get();
      bindMethods(this.model, HELPER_METHODS_TO_BIND);
      this.autorunRender();
      this.autorunSubscriptions();

      for (var _len2 = arguments.length, args = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        args[_key2] = arguments[_key2];
      }

      if (oldComponentWillMount) oldComponentWillMount.call.apply(oldComponentWillMount, [this].concat(args));
      this.__ranComponentWillMount = true;
      // run componentDidSubscribe custom lifecycle hook here if subscriptions
      // finished synchronously
      if (this.__loaded && this.componentDidSubscribe) {
        this.componentDidSubscribe();
      }
    },

    // TODO: Implement queueing
    componentWillReceiveProps: function componentWillReceiveProps() {
      var _this = this;

      for (var _len3 = arguments.length, args = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
        args[_key3] = arguments[_key3];
      }

      return (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee() {
        var nextProps, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, dataFn;

        return _regenerator2.default.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                nextProps = args[0];
                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context.prev = 4;
                _iterator = (0, _getIterator3.default)(_this.__dataFns);

              case 6:
                if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                  _context.next = 15;
                  break;
                }

                dataFn = _step.value;
                _context.next = 10;
                return dataFn(nextProps);

              case 10:
                if (!_this.unmounted) {
                  _context.next = 12;
                  break;
                }

                return _context.abrupt('return');

              case 12:
                _iteratorNormalCompletion = true;
                _context.next = 6;
                break;

              case 15:
                _context.next = 21;
                break;

              case 17:
                _context.prev = 17;
                _context.t0 = _context['catch'](4);
                _didIteratorError = true;
                _iteratorError = _context.t0;

              case 21:
                _context.prev = 21;
                _context.prev = 22;

                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }

              case 24:
                _context.prev = 24;

                if (!_didIteratorError) {
                  _context.next = 27;
                  break;
                }

                throw _iteratorError;

              case 27:
                return _context.finish(24);

              case 28:
                return _context.finish(21);

              case 29:
                if (oldComponentWillReceiveProps) {
                  oldComponentWillReceiveProps.call.apply(oldComponentWillReceiveProps, [_this].concat(args));
                }

              case 30:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, _this, [[4, 17, 21, 29], [22,, 24, 28]]);
      }))();
    },
    componentWillUnmount: function componentWillUnmount() {
      for (var _len4 = arguments.length, args = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
        args[_key4] = arguments[_key4];
      }

      if (oldComponentWillUnmount) oldComponentWillUnmount.call.apply(oldComponentWillUnmount, [this].concat(args));
      this.unmounted = true;
      // Stop render computation
      (0, _observerUtil.unobserve)(this.render);
      // Stop all subscription params computations
      for (var index = 0; index < this.__dataFns.length; index++) {
        var computationName = getComputationName(index);
        this.__comps[computationName] && (0, _observerUtil.unobserve)(this.__comps[computationName]);
        delete this.__comps[computationName];
      }
      delete this.__dataFns;
      // Destroy all subscription items
      for (var key in this.__items) {
        this.__destroyItem(key, true);
      }
      this.model.destroy();
      delete this.$scope; // delete model alias
      delete this.model; // delete the actual model
    },
    autorunRender: function autorunRender() {
      var _this2 = this;

      var oldRender = this.render;
      var loadingRenderWrapper = function loadingRenderWrapper() {
        _this2.__rendered = true;
        if (_this2.__loaded) {
          return oldRender.call(_this2);
        } else {
          // When in React Native env, don't use any loading spinner
          if (typeof navigator !== 'undefined' && navigator.product === 'ReactNative') {
            return null;
          } else {
            return _react2.default.createElement('div', { className: 'Loading' });
          }
        }
      };
      var fn = _lodash2.default.debounce(function () {
        if (_this2.unmounted) return;
        _this2.setState(DUMMY_STATE);
      });
      this.render = (0, _observerUtil.observe)(loadingRenderWrapper, {
        scheduler: fn,
        lazy: true
      });
    },

    // TODO: When we change subscription params quickly, we are going to
    //       receive a race condition when earlier subscription result might
    //       take longer to process compared to the same newer subscription.
    //       Implement Queue.
    autorunSubscriptions: function autorunSubscriptions() {
      var _this3 = this;

      return (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee3() {
        var _loop, index, _ret;

        return _regenerator2.default.wrap(function _callee3$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _this3.__items = {};
                _this3.__comps = {};
                _this3.__dataFns = [];
                _loop = /*#__PURE__*/_regenerator2.default.mark(function _loop(index) {
                  var fn, subscriptions, dataFn;
                  return _regenerator2.default.wrap(function _loop$(_context3) {
                    while (1) {
                      switch (_context3.prev = _context3.next) {
                        case 0:
                          fn = fns[index];
                          subscriptions = {};

                          dataFn = function () {
                            var _ref = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee2(props) {
                              var prevSubscriptions, computationName, subscribeFn, computation, keys, promises, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, key;

                              return _regenerator2.default.wrap(function _callee2$(_context2) {
                                while (1) {
                                  switch (_context2.prev = _context2.next) {
                                    case 0:
                                      prevSubscriptions = subscriptions || {};
                                      computationName = getComputationName(index);

                                      subscribeFn = function subscribeFn() {
                                        subscriptions = fn.call(_this3, props);
                                      };

                                      computation = (0, _observerUtil.observe)(subscribeFn, {
                                        scheduler: dataFn
                                      });

                                      _this3.__comps[computationName] = computation;

                                      keys = _lodash2.default.union(_lodash2.default.keys(prevSubscriptions), _lodash2.default.keys(subscriptions));

                                      keys = _lodash2.default.uniq(keys);
                                      promises = [];
                                      _iteratorNormalCompletion2 = true;
                                      _didIteratorError2 = false;
                                      _iteratorError2 = undefined;
                                      _context2.prev = 11;
                                      _iterator2 = (0, _getIterator3.default)(keys);

                                    case 13:
                                      if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
                                        _context2.next = 28;
                                        break;
                                      }

                                      key = _step2.value;

                                      if (_lodash2.default.isEqual(subscriptions[key], prevSubscriptions[key])) {
                                        _context2.next = 25;
                                        break;
                                      }

                                      if (!subscriptions[key]) {
                                        _context2.next = 24;
                                        break;
                                      }

                                      _context2.t0 = promises;
                                      _context2.next = 20;
                                      return _this3.__initItem(key, subscriptions[key]);

                                    case 20:
                                      _context2.t1 = _context2.sent;

                                      _context2.t0.push.call(_context2.t0, _context2.t1);

                                      _context2.next = 25;
                                      break;

                                    case 24:
                                      _this3.__destroyItem(key, true);

                                    case 25:
                                      _iteratorNormalCompletion2 = true;
                                      _context2.next = 13;
                                      break;

                                    case 28:
                                      _context2.next = 34;
                                      break;

                                    case 30:
                                      _context2.prev = 30;
                                      _context2.t2 = _context2['catch'](11);
                                      _didIteratorError2 = true;
                                      _iteratorError2 = _context2.t2;

                                    case 34:
                                      _context2.prev = 34;
                                      _context2.prev = 35;

                                      if (!_iteratorNormalCompletion2 && _iterator2.return) {
                                        _iterator2.return();
                                      }

                                    case 37:
                                      _context2.prev = 37;

                                      if (!_didIteratorError2) {
                                        _context2.next = 40;
                                        break;
                                      }

                                      throw _iteratorError2;

                                    case 40:
                                      return _context2.finish(37);

                                    case 41:
                                      return _context2.finish(34);

                                    case 42:
                                      _context2.next = 44;
                                      return _promise2.default.all(promises);

                                    case 44:
                                    case 'end':
                                      return _context2.stop();
                                  }
                                }
                              }, _callee2, _this3, [[11, 30, 34, 42], [35,, 37, 41]]);
                            }));

                            return function dataFn(_x) {
                              return _ref.apply(this, arguments);
                            };
                          }();

                          _this3.__dataFns.push(dataFn);
                          _context3.next = 6;
                          return dataFn(_this3.props);

                        case 6:
                          if (!_this3.unmounted) {
                            _context3.next = 8;
                            break;
                          }

                          return _context3.abrupt('return', {
                            v: void 0
                          });

                        case 8:
                        case 'end':
                          return _context3.stop();
                      }
                    }
                  }, _loop, _this3);
                });
                index = 0;

              case 5:
                if (!(index < fns.length)) {
                  _context4.next = 13;
                  break;
                }

                return _context4.delegateYield(_loop(index), 't0', 7);

              case 7:
                _ret = _context4.t0;

                if (!((typeof _ret === 'undefined' ? 'undefined' : (0, _typeof3.default)(_ret)) === "object")) {
                  _context4.next = 10;
                  break;
                }

                return _context4.abrupt('return', _ret.v);

              case 10:
                index++;
                _context4.next = 5;
                break;

              case 13:
                _this3.__loaded = true;
                // run componentDidSubscribe custom lifecycle hook here if subscriptions
                // finished Asynchronously (after componentWillMount already executed)
                if (_this3.__ranComponentWillMount && _this3.componentDidSubscribe) {
                  _this3.componentDidSubscribe();
                }
                // Sometimes all the subscriptions might go through synchronously
                // (for example if we are only subscribing to local data).
                // In this case we don't need to manually trigger forceUpdate
                // since render will execute on its own later in the lifecycle.
                if (_this3.__rendered) _this3.setState(DUMMY_STATE);

              case 16:
              case 'end':
                return _context4.stop();
            }
          }
        }, _callee3, _this3);
      }))();
    },

    // TODO: Maybe implement queueing. Research if race condition is present.
    __initItem: function __initItem(key, params) {
      var _this4 = this;

      return (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee4() {
        var constructor, item, keyModelName;
        return _regenerator2.default.wrap(function _callee4$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                constructor = getItemConstructor(params);
                item = new constructor(_this4.model, key, params);
                _context5.next = 4;
                return item.init();

              case 4:
                if (!_this4.unmounted) {
                  _context5.next = 6;
                  break;
                }

                return _context5.abrupt('return', item.destroy());

              case 6:
                if (_this4.__items[key]) _this4.__destroyItem(key);
                // Expose scoped model under the same name with prepended $
                keyModelName = getScopedModelName(key);

                if (!_this4[keyModelName]) _this4[keyModelName] = _this4.model.at(key);
                item.refModel();
                _this4.__items[key] = item;

              case 11:
              case 'end':
                return _context5.stop();
            }
          }
        }, _callee4, _this4);
      }))();
    },
    __destroyItem: function __destroyItem(key, terminate) {
      if (!this.__items[key]) return console.error('Trying to destroy', key);
      this.__items[key].unrefModel();
      var keyModelName = getScopedModelName(key);
      if (terminate) delete this[keyModelName];
      this.__items[key].destroy();
      delete this.__items[key];
    },
    __removeItemRefs: function __removeItemRefs(key) {
      this.__items[key].unrefModel();
    }
  };
};

function generateScopedModel() {
  var path = DEFAULT_COLLECTION + '.' + _model2.default.id();
  return _model2.default.scope(path);
}

function isExtraQuery(queryParams) {
  return queryParams.$count || queryParams.$aggregate;
}

function getItemConstructor(subscription) {
  if (typeof subscription === 'string') return _Local2.default;

  var _subscription = (0, _slicedToArray3.default)(subscription, 2),
      params = _subscription[1];

  return typeof params === 'string' || !params ? _Doc2.default : isExtraQuery(params) ? _QueryExtra2.default : _Query2.default;
}

function getComputationName(index) {
  return '' + SUBSCRIBE_COMPUTATION_NAME + index;
}

function bindMethods(object, methodsToBind) {
  var _iteratorNormalCompletion3 = true;
  var _didIteratorError3 = false;
  var _iteratorError3 = undefined;

  try {
    for (var _iterator3 = (0, _getIterator3.default)(methodsToBind), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
      var method = _step3.value;

      object[method] = object[method].bind(object);
    }
  } catch (err) {
    _didIteratorError3 = true;
    _iteratorError3 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion3 && _iterator3.return) {
        _iterator3.return();
      }
    } finally {
      if (_didIteratorError3) {
        throw _iteratorError3;
      }
    }
  }
}

function getScopedModelName(key) {
  return '$' + key;
}
module.exports = exports['default'];